/******************************************************************************
***
***				COPYRIGHT
***
***	Die Daten in dieser Datei sind Eigentum der 
***
***	YARD Software GmbH, 51107 Koeln, Deutschland.
***
***	Diese Datei gilt als Werk im Sinne des Urheberrechtsgesetzes und ist
***	damit urheberrechtlich geschuetzt. Jede Verwertung, insbesondere die
***	Vervielfaeltigung, Aenderung und Weitergabe, auch auszugsweise,
***	ist nur mit ausdruecklicher schriftlicher Genehmigung der YARD
***	Software GmbH zulaessig.
***
***	Ebenso ist die Speicherung dieser Datei auf Datentraegern wie z.B.
***	Festplatten, Disketten, Magnetbaendern, CDs und aehnlichen, auch 
***	auszugsweise, nur mit ausdruecklicher schriftlicher Genehmigung der 
***	YARD Software GmbH zulaessig.
***
***	Zuwiderhandlungen unterliegen den Strafbestimmungen des Urheberrechts-
***	gesetzes.
***
******************************************************************************/

static char *demo1_ec_sccsid = "@(#) demo4.ec 1.1 98/11/05 16:44:29";

/*
**	This program shows the use of the dynamic SQL statements
**	PREPARE and DESCRIBE to process an unknown SELECT statement
**	at runtime.
**	
**	PREREQUISITES:
**	
**	The database RENTAL must be installed using 'yarddemo'
**	The environment variable YARDDBS must be set.
*/

#include <stdio.h>

#include "sqltypes.h"
#include "sqlblob.h"

/*
**	Function prototypes
*/
static void setup_sqlda(sqlda_t *da);
static void print_column_header(sqlda_t *da);
static void print_column_data(sqlda_t *da);

main(int argc, char **argv)
{
/*
**	Host variable to hold the
**	select statement.
*/
exec sql begin declare section;
	char selstmt[512];
exec sql end declare section;

	int rowcnt = 0;

	/*
	**	sqlda structure to hold the SELECT
	**	result.
	*/
	sqlda_t *selda;

	/*
	**	The program will stop in case of an error
	*/
	exec sql whenever sqlerror stop;

	/*
	**	Connect to the database server with user DBA
	**	The DBS name is taken from the environment variable
	**	YARDDBS during connect.
	*/
	exec sql connect user dba;

	/*
	**	Open the database RENTAL
	*/
	exec sql database rental;

	/*
	**	Store the SELECT statement into a host
	**	variable.
	**	Assume the SELECT is unknown, for example
	**	it can be given at runtime trough user
	**	input.
	*/
	sprintf(selstmt, "select * from rental");

	/*
	**	Step 1:
	**	The first step is to PREPARE the statement.
	**	The statement will be checked by the database
	**	server.
	**	If any error occurs (eg. unknown column, table)
	**	the program stops.
	*/
	exec sql prepare dyn_stmt from :selstmt;

	/*
	**	Step 2:
	**	DESCRIBE the statement in the second step.
	**	We will get an sqlda structure which holds
	**	information (column name, data type, etc.)
	**	about every column of the SELECT list.
	*/
	exec sql describe dyn_stmt into selda;

	/*
	**	Step 3:
	**	Now we can use the sqlda structure to
	**	allocate memory for the SELECT results.
	*/
	setup_sqlda(selda);

	/*
	**	Step 4:
	**	DECLARE a cursor to SELECT all customers, 
	**	which currently have rent a car.
	*/
	exec sql declare dyn_cursor cursor for dyn_stmt;

	/*
	**	Step 5:
	**	Open the cursor
	*/
	exec sql open dyn_cursor;

	/*
	**	Print a header with the column names
	*/
	print_column_header(selda);

	/*
	**	Step 6:
	**	Fetch the cursor until no more
	**	rows (SQLCODE == SQLNOTFOUND)
	**	will be retrived.
	*/
	for (;;)
	{
		exec sql fetch dyn_cursor into descriptor selda;
		if (SQLCODE == SQLNOTFOUND)
			break;
		rowcnt++;

		/*
		**	Print the data
		*/
		print_column_data(selda);
		printf("\n");
	}

	printf("\n%d rows retrieved\n", rowcnt);

	/*
	**	Step 7:
	**	Close the cursor
	*/
	exec sql close dyn_cursor;

	/*
	**	Close the database
	*/
	exec sql close database;

	/*
	**	Disconnect from database server
	*/
	exec sql disconnect;

	exit(0);
}

/*
**	Allocate memory according to the data type
**	for the data element of each sqlha structure.
**	The number of sqlha structures corresponds
**	to the number of select elements.
**	The element hostcnt of the sqlda structure
**	holds the number of sqlha structures.
**	NOTE: We don't ask for the return code of
**	      calloc() in this example.
*/
static void setup_sqlda(sqlda_t *da)
{
	register sqlha_t *col;
	register int i;
	int memsize;

	for (i = 0; i < da->hostcnt; i++)
	{
		/*
		**	Store the address of the
		**	current sqlha structure
		**	for easier handling.
		*/
		col = &da->sqlha[i];

		/*
		**	Get the appropiate memory size
		**	to hold the data for a column.
		*/
		switch(col->type)
		{
			case SQL_VARCHAR:
			case SQL_CHAR:
				/*
				**	We want only the first
				**	19 bytes of char data.
				**	So we set length to 20
				**	(19 and 1 for the 0 byte).
				*/
				col->length = 20;
				memsize = col->length;
				break;
			case SQL_INTEGER:
			case SQL_SERIAL:
				memsize = sizeof(long);
				break;
			case SQL_SMALLINT:
				memsize = sizeof(short);
				break;
			case SQL_DOUBLE:
			case SQL_FLOAT:
				memsize = sizeof(double);
				break;
			case SQL_REAL:
				memsize = sizeof(float);
				break;
			case SQL_DATE:
				/*
				**	We want the date type
				**	to be converted to char.
				**	So the type field is set
				**	to char and length to 11
				**	to store a value like:
				**	<DD.MM.YYYY\0>
				*/
				col->type = SQL_CHAR;
				col->length = 11;
				memsize = col->length;
				break;
			case SQL_NUMERIC:
			case SQL_DECIMAL:
			case SQL_MONEY:
				/*
				**	All decimal data is converted
				**	to char. The maximal length is
				**	set to 20 (incl. the 0 byte).
				*/
				col->type = SQL_CHAR;
				col->length = col->precision + 10;
				memsize = col->length;
				break;
			case SQL_BLOB:
			case SQL_TEXT:
				/*
				**	The blob structure is a special
				**	structure to hold blob values.
				**	It is declared in sqlblob.h
				*/
				memsize = sizeof(blob_t);
				break;
			default:
				fprintf(stderr, "Unknown data type: %d ",
					col->type);
				fprintf(stderr, "column: %s\n",
					col->colname);
				exit(1);
		}

		/*
		**	Allocate memory for data pointer
		*/
		col->data = (char *)calloc(1, memsize);

		/*
		**	If the column allows null values,
		**	allocate memory for the indicator
		**	pointer.
		*/
		if (col->nulls == 1)
		{
			col->indicator = (short *)malloc(sizeof(short));
		}

		/*
		**	If the type is BLOB or TEXT, initialize
		**	the BLOB structure so that only the size
		**	of the data is retrieved during fetch time
		*/
		if (col->type == SQL_BLOB || col->type == SQL_TEXT)
		{
			((blob_t *)col)->b_type = BLOB_SIZE;
		}
	}

	return;
}


/*
**	For each column print the name
**	(element colname of the sqlha
**	structure) to stdout.
*/
static void print_column_header(sqlda_t *da)
{
	register int i;

	for (i = 0; i < da->hostcnt; i++)
	{
		printf("%-20.20s ", da->sqlha[i].colname);
	}
	printf("\n\n");

	return;
}

/*
**	For each column print the name
**	(element colname of the sqlha
**	structure) to stdout.
*/
static void print_column_data(sqlda_t *da)
{
	register sqlha_t *col;
	register int i;
	int memsize;
	char buf[32];

	for (i = 0; i < da->hostcnt; i++)
	{
		col = &da->sqlha[i];

		/*
		**	Ask for a null value. In this
		**	case the indicator pointer holds
		**	the value -1.
		*/
		if (col->nulls && *col->indicator == -1)
		{
			printf("%-20.20s ", "NULL Value");
			continue;
		}

		/*
		**	Print the data according the the type
		*/
		switch(col->type)
		{
			case SQL_VARCHAR:
			case SQL_CHAR:
			case SQL_DATE:
			case SQL_NUMERIC:
			case SQL_DECIMAL:
			case SQL_MONEY:
				printf("%-20.20s ", col->data);
				break;
			case SQL_INTEGER:
			case SQL_SERIAL:
				printf("%-20ld ", *(long *)col->data);
				break;
			case SQL_SMALLINT:
				printf("%20d ", *(short *)col->data);
				break;
			case SQL_DOUBLE:
			case SQL_FLOAT:
				printf("%20f ", *(double *)col->data);
				break;
			case SQL_REAL:
				printf("%20f ", *(float *)col->data);
				break;
			case SQL_BLOB:
				sprintf(buf, "<BLOB> (%d)", 
					((blob_t *)col)->b_size);
				break;
			case SQL_TEXT:
				sprintf(buf, "<TEXT> (%d)", 
					((blob_t *)col)->b_size);
				break;
		}

	}

	return;
}
