
{
	package DBD::YARD;
	require DBI;
	require DynaLoader;
	@ISA = qw(DynaLoader);

	bootstrap DBD::YARD;

	$err = 0;	# holds error code   for DBI::err
	$errstr = "";	# holds error string for DBI::errstr
	$drh = undef;	# holds driver handle once initialised

	sub driver
	{
		my($class, $attr) = @_;

		$class .= "::dr";

		warn "YARDHOME not set!\n" unless $ENV{YARDHOME};

		$drh = DBI::_new_drh($class, {
		'Name'                  => 'YARD',
		'Version'               => '4.4',
		'Err'                   => \$DBD::YARD::err,
		'Errstr'                => \$DBD::YARD::errstr,
		'State'                 => \$DBD::YARD::state,
		'Attribution' 		=> 'DBD::YARD '
		});

		DBD::YARD::dr::init($drh);

		$drh;
	}

	1;
}

{
	package DBD::YARD::dr;
	use strict;

	sub errstr
	{
		DBD::YARD::errstr(@_);
	}

	sub connect
	{
		my ($drh, $dbname, $dbuser, $dbpass) = @_;

		$dbname = "" unless(defined $dbname);
		$dbuser = "" unless(defined $dbuser);
		$dbpass = "" unless(defined $dbpass);

		# Create new database connection handle for driver
		my $dbh = DBI::_new_dbh($drh, {
				'Name' => $dbname,
				'User' => $dbuser,
				'Pass' => $dbpass
			});

		# Initialize database connection
		DBD::YARD::db::connect($dbh, $dbname, $dbuser, $dbpass)
			or return undef;

		$dbh;
	}
	1;
}

{
	package DBD::YARD::db;
	use strict;

	sub errstr
	{
		DBD::YARD::errstr(@_);
	}

	sub prepare
	{
		my($dbh, $statement)= @_;

		my $sth = DBI::_new_sth($dbh, {
			'Statement' => $statement,
			});

		DBD::YARD::st::prepare($sth, $statement)
			or return undef;

		$sth;
	}

	1;
}

{
	package DBD::YARD::st;
	use strict;

	sub errstr
	{
		DBD::YARD::errstr(@_);
	}
	1;
}

__END__

=head1 NAME

DBD::YARD - Access to YARD Databases

=head1 SYNOPSIS

use DBI;

=head1 DESCRIPTION

This document describes DBD::YARD version 4.4.

You should also read the documentation for DBI as this document
qualifies what is stated there.

This document covers parts of DBI and most of DBD::YARD.
The DBI document's comments
about DBI should take precedence over comments in this document (but
this document may be a better reflection of the actual behaviour of
DBD::YARD).


=head1 Loading DBD::YARD

To use the DBD::YARD software, you need to load the DBI software.

    use DBI;

Under normal circumstances, you should then connect to your database
using the notation in the section "CONNECTING TO A DBS" which
calls DBI->connect().

=head2 Driver Attributes and Methods

If you have a burning desire to do so, you can explicitly install the
YARD driver independently of connecting to any database using:

    $drh = DBI->install_driver('YARD');

This gives you a reference to the driver, aka the driver handle.
If the load fails, your program stops immediately (unless, perhaps,
you eval the statement).

Once you have the driver handle, you can interrogate the driver for some
basic information:

    print "Driver Information\n";
    # Type is always 'dr'.
    print "    Type:                  $drh->{Type}\n";
    # Name is always 'YARD'.
    print "    Name:                  $drh->{Name}\n";
    # Version is the version of DBD::YARD (eg 0.51).
    print "    Version:               $drh->{Version}\n";

Once you have the driver loaded, you can connect to a database, or you
can sever all connections to databases with disconnect_all.

    $drh->disconnect_all;

=head1 CONNECTING TO A DBS

Read the DBI specification for information about "dbi:YARD:" as
the initial portion of the connect string.

To connect to a DBS, you can use the connect function, which
yields a valid reference or database handle if it is successful.
If the driver itself cannot be loaded (by the DBI->install_driver()
method mentioned above), DBI aborts the script (and DBD::YARD can
do nothing about it because it wasn't loaded successfully).

    $dbh = DBI->connect("dbi:YARD:$dbs", $user, $pass);
    $dbh = DBI->connect("dbi:YARD:$dbs", $user);

The string following the second colon is all that is passed to the
DBD::YARD code. It must have the same format as the environment variable
YARDDBS (dbs@system). All connections run through the net interface 
(even local), so yardnetd must be running.
Note that if you specify the username but not the password, the
password '' will be used.
You cannot specify certain attributes in the connect call. By default,
Autocommit is enabled for a connection.

Once you have a database handle, you can interrogate it for some basic
information about the connection, etc.

     print "Database Information\n";
     # Type is always 'db'
     print "    Type:                    $dbh->{Type}\n";
     # Name is the name of the DBS specified at connect
     print "    DBS Name:           $dbh->{Name}\n";
     # AutoCommit is 1 (true) if the database commits each statement.
     print "    AutoCommit:              $dbh->{AutoCommit}\n";

If $dbh->{PrintError} is true, then DBI will report each error
automatically on STDERR when it is detected.
The error is also available via the package variables $DBI::errstr and
$DBI::err.

If $dbh->{PrintError} is false, then DBI does not report any errors
when it detects them; it is up to the user to note that errors have
occurred and to report them.

Note that connecting to a DBS doesn't open a database (as usual when using
YARD-SQL).


=head1 DISCONNECTING FROM A DBS

You can also disconnect from the DBS:

    $dbh->disconnect;

This will rollback any uncommitted work. Close the database by executing
'CLOSE DATABASE' before disconnecting.
Note that this does not destroy the database handle.
You need to do an explicit 'undef $dbh' to destroy the handle.
Any statements prepared using this handle are finished (see below) and
cannot be used again.
All space associated with the statements is released.

=head1 SIMPLE STATEMENTS

Given a database connection, you can execute a variety of simple
statements using a variety of different calls:

    $dbh->commit;
    $dbh->rollback;

These two operations commit or rollback the current transaction.
If AutoCommit is set to 1, then they do nothing useful.
If AutoCommit is set to 0, then a new transaction is started.

You can execute most preparable parameterless statements using:

    $dbh->do($stmt);

The statement should not be either SELECT (other than SELECT...INTO
TEMP) or CALL.

You can execute an arbitrary statement with parameters using:

    $dbh->do($stmt, @parameters);
    $dbh->do($stmt, $param1, $param2);

Again, the statement must not be a SELECT or CALL.
The values in @parameters (or the separate values) are bound to the
question marks in the statement string.

    $sth = $dbh->prepare($stmt);
    $sth->execute(@parameters);

You can embed an arbitrary string inside a statement with any quote
marks correctly handled by invoking:

    $dbh->quote($string);

This method is provided by the DBI package implementation and is
inherited by the DBD::YARD package.
The string is enclosed in single quotes, and any embedded single
quotes are doubled up, which conforms to the SQL-92 standard.
This would typically be used in a context such as:

    $value = "Doesn't work unless quotes (\"'\" and '\"') are handled";

    $stmt = "INSERT INTO SomeTable(SomeColumn) " .
            "VALUES(" . $dbh->quote($value) . ")";

Doing this ensures that the data in $values will be interpreted
correctly, regardless of what quotes appear in $value (unless it
contains newline characters).
Note that the alternative assignment below does not work!

    $stmt = "INSERT INTO SomeTable(SomeColumn) VALUES($dbh->quote($value))";

=head1 CREATING STATEMENTS

You can also prepare a statement for multiple uses, and you can do
this for SELECT and CALL statements which return data
(cursory statements) as well as non-cursory statements which return no
data.
You create a statement handle (another reference) using:

    $sth = $dbh->prepare($stmt);

If the statement is a SELECT which returns data (not SELECT...INTO TEMP) or
an CALL for a procedure which returns values, then a cursor is
declared for the prepared statement.

According to the DBI specification, the prepare call accepts an
optional attributes parameter which is a reference to a hash.
At the moment, no parameters are recognized.


Once the statement is prepared, you can execute it:

    $sth->execute;

For a non-cursory statement, this simply executes the statement.
If the statement is executed successfully, then the number of rows
affected will be returned.
If an error occurs, the returned value will be undef.
If the statement does not affect any rows, the string returned is
"0E0" which evaluates to true but also to zero.

For a cursory statement, it opens the cursor.
If the cursor is opened successfully, then it returns the value "0E0"
which evaluates to true but also to zero.
If an error occurs, the returned value will be undef.


If the statement is a cursory statement, you can retrieve the values
in any of a number of ways, as described in the DBI specification.

    $ref = $sth->fetch;
    @row = @{$ref};

    @row = @{$sth->fetch};  # Shorthand for above...

    @row = $sth->fetchrow;

As usual, you have to worry about whether this worked or not.  You would
normally, therefore, use:

    while ($ref = $sth->fetch)
    {
        # We know we got some data here
        ...
    }
    # Investigate whether an error occurred or the SELECT
    # simply had nothing more to return.
    if (length($sth->errstr) > 0)
    {
        # Process error...
    }

The returned data includes blobs mapped into strings.
Note that byte blobs might contain ASCII NUL '\0' characters.
Perl knows how long the strings are and does preserve NUL in the
middle of a byte blob.
However, you may need to be careful deciding how to handle this
string.


When you have fetched as many rows as required, you close the cursor using:

    $sth->finish;

This simply closes the cursor; it does not free the cursor or the statement.
That is done when you destroy (undef) the statement handle:

    undef $sth;

You can also implicitly rebind a statement handle to a new statement
by simply using the same variable again.
This does not cause any memory leaks.


=head1 KNOWN RESTRICTIONS

=item *
The method data_sources is not implemented.

=item *
The attribute LongTruncOk is ignored.
Blobs might be truncated silently.


=head1 SEE ALSO

perl(1), perldoc for DBI.

YARD-SQL and YARD-ADDONS Users and Reference Guide, available as HTML document from www.yard.de.

=head1 AUTHOR

DBD::YARD by Thomas Schonhoven, YARD Software GmbH.

=head1 COPYRIGHT

The DBD::YARD module is Copyright (c) 1997,1998 Thomas Schonhoven. Germany.
The DBD::YARD module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=cut
