/**
 * You may distribute under the terms of either the GNU General Public
 * License or the Artistic License, as specified in the Perl README file.
 *
 */

/******************************************************************************
***
***	YARD interface for Perl
***
******************************************************************************/

#include <stdarg.h>

#include "clibasic.h"
#include "sql.h"
#include "sqlext.h"
#include "dbdimp.h"

#define	ERRBUFSIZE		512
#define	GETDATABUFSIZE		32768
#define	MAXERRORSTACKSIZE	10

#define ENV_HANDLE		0x001
#define CON_HANDLE		0x002
#define STMT_HANDLE		0x003

static char logfilename[1024];
static char envstring[1024];
static HENV henv = NULL;
static char errorstack[MAXERRORSTACKSIZE][ERRBUFSIZE+1+20];
static int errorstackpos;

static SV *dbd_errnum = NULL;
static SV *dbd_errstr = NULL;
static SV *dbd_state = NULL;

static int yard_clifailed(int,void*,int);
static void *yard_str2handle(char *);
static char *yard_handle2str(void*);
static void yard_log(int, char*,...);
static int yard_pusherror(char*,char*,int);
static void yard_clearerrorstack();

DBISTATE_DECLARE;


/*
void
dbd_init(dbistate)
    dbistate_t *dbistate;
{
    DBIS = dbistate;
    dbd_errnum = GvSV(gv_fetchpv("DBD::YARD::err",    1, SVt_IV));
    dbd_errstr = GvSV(gv_fetchpv("DBD::YARD::errstr", 1, SVt_PV));
}
*/

void
dbd_dr_init(dbistate)
dbistate_t *dbistate;
{
	DBIS = dbistate;
        dbd_errnum = GvSV(gv_fetchpv("DBD::YARD::err", 1, SVt_IV));
        dbd_errstr = GvSV(gv_fetchpv("DBD::YARD::errstr", 1, SVt_PV));
        dbd_state  = GvSV(gv_fetchpv("DBD::YARD::state", 1, SVt_PV));
}

/******************************************************************************
***
***	Initialisation function. This function must be called before any other
***	function is called.
***	If this function fails, additional error information might be 
***	displayed on stderr.
***
***	Parameter:	No
***
***	Returnvalue:	1 on success
***			0 on error
***
******************************************************************************/

int dbd_dr_driver( SV *drh )
{
	int rc;

	D_imp_drh(drh);

	yard_log(1,"yard_init: YARDHOME = #%s#\n", getenv("YARDHOME"));

	rc = SQLAllocEnv(&henv);
	if (rc)
	{
		yard_pusherror("SQLAllocEnv() failed","XXXXX",-1);
		return 0;
	}
	return 1;
}

/******************************************************************************
***
***	Connect to a DBS/database
***
***	Parameter:	DBS (yarddbs@system)
***			userid
***			password
***
***	Returnvalue:	1 on success
***			0 on error
***
******************************************************************************/

int dbd_db_connect( SV *dbh, char *dbs, char *user, char *passwd )
{
	int rc;
	char dsn[100];
	
	D_imp_dbh(dbh);

	yard_log(1,"connect: dbs = #%s#, userid = #%s#, password %sspecified, hdbc = 0x%08x\n",dbs,user,passwd[0]?"":"not ",imp_dbh->hdbc);
	yard_clearerrorstack();

	rc = SQLAllocConnect(henv,&imp_dbh->hdbc);
	if (yard_clifailed(rc,henv,ENV_HANDLE))
		return 0;

	strcpy(dsn,dbs);
	rc = SQLConnect(imp_dbh->hdbc,dsn,strlen(dsn),user,strlen(user),passwd,
		strlen(passwd));
	if (yard_clifailed(rc,imp_dbh->hdbc,CON_HANDLE))
	{
		SQLFreeConnect(imp_dbh->hdbc);
		imp_dbh->hdbc = NULL;
		return 0;
	}
	DBIc_on(imp_dbh,DBIcf_IMPSET);    /* imp_dbh set up now             */
	DBIc_on(imp_dbh,DBIcf_ACTIVE);    /* call disconnect before freeing */
	DBIc_on(imp_dbh,DBIcf_AutoCommit);/* This is the default	    */

	/* should be ODBC default !! */

	rc = SQLSetConnectOption(imp_dbh->hdbc, SQL_AUTOCOMMIT,1);
	if (yard_clifailed(rc,imp_dbh->hdbc,CON_HANDLE))
		return 0;

	return 1;
}

int dbd_db_commit(SV *dbh)
{
	int rc;
	D_imp_dbh(dbh);

	yard_log(1,"Commit\n");

	rc = SQLTransact(henv,imp_dbh->hdbc,SQL_COMMIT);
	if (yard_clifailed(rc,imp_dbh->hdbc,CON_HANDLE))
		return 0;
	return 1;
}

int dbd_db_rollback(SV *dbh)
{
	int rc;
	D_imp_dbh(dbh);

	yard_log(1,"Rollback\n");

	rc = SQLTransact(henv,imp_dbh->hdbc,SQL_ROLLBACK);
	if (yard_clifailed(rc,imp_dbh->hdbc,CON_HANDLE))
		return 0;
	return 1;
}

/******************************************************************************
***
***
***
******************************************************************************/

int dbd_db_disconnect(SV *dbh)
{
	int rc;
	D_imp_dbh(dbh);

	yard_log(1,"Disconnect\n");

	rc = SQLDisconnect(imp_dbh->hdbc);
	if (yard_clifailed(rc,imp_dbh->hdbc,CON_HANDLE))
		return 0;
	DBIc_off(imp_dbh,DBIcf_ACTIVE);
	return 1;
}

/******************************************************************************
***
***
***
******************************************************************************/

void dbd_db_destroy(SV *dbh)
{
	D_imp_dbh(dbh);

	yard_log(1,"db_destroy\n");

	if (DBIc_is(imp_dbh,DBIcf_ACTIVE))
		dbd_db_disconnect(dbh);
	DBIc_off(imp_dbh,DBIcf_IMPSET);
}

/******************************************************************************
***
***
***
******************************************************************************/

int dbd_db_STORE(SV *dbh, SV *keysv, SV *valuesv)
{
	int rc;
	D_imp_dbh(dbh);
	STRLEN keyLength;
	char *key = SvPV( keysv, keyLength );
	SV *retsv = NULL;
	int on = SvTRUE(valuesv);

	yard_log(1,"db_STORE: key = %s\n", key);

	if (!strcmp(key,"AutoCommit"))
	{
		yard_log(1,"Setting autocommit to %d\n",on);
		if (on)
		{
			rc = SQLSetConnectOption(imp_dbh->hdbc,
				SQL_AUTOCOMMIT,1);
			if (yard_clifailed(rc,imp_dbh->hdbc,CON_HANDLE))
				return 0;
			DBIc_on(imp_dbh,DBIcf_AutoCommit);
		}
		else
		{
			rc = SQLSetConnectOption(imp_dbh->hdbc,
				SQL_AUTOCOMMIT,0);
			if (yard_clifailed(rc,imp_dbh->hdbc,CON_HANDLE))
				return 0;
			DBIc_off(imp_dbh,DBIcf_AutoCommit);
		}
	}
	return TRUE;
}

/******************************************************************************
***
***
***
******************************************************************************/

SV *dbd_db_FETCH(SV *dbh, SV *keysv)
{
	D_imp_dbh(dbh);
	STRLEN keyLength;
	char *key = SvPV( keysv, keyLength );
	SV *retsv = NULL;

	yard_log(1,"db_FETCH: key = %s\n", key);

	if (!strcmp(key,"AutoCommit"))
		retsv = newSViv((DBIc_has(imp_dbh,DBIcf_AutoCommit))?1:0);
	if (!retsv)
		return Nullsv;

	return sv_2mortal( retsv );
}

/******************************************************************************
***
***
***
******************************************************************************/

int dbd_st_prepare(SV *sth, char *stmt, SV *attribs)
{
	int rc;

	D_imp_sth(sth);
	D_imp_dbh_from_sth;

	yard_log(1,"prepare: stmt = #%s#\n",stmt);
	yard_clearerrorstack();

	rc = SQLAllocStmt(imp_dbh->hdbc,&imp_sth->hstmt);
	if (yard_clifailed(rc,imp_dbh->hdbc,CON_HANDLE))
		return 0;
	rc = SQLPrepare(imp_sth->hstmt,stmt,strlen(stmt));
	if (yard_clifailed(rc,imp_sth->hstmt,STMT_HANDLE))
	{
		SQLFreeStmt(imp_sth->hstmt,SQL_DROP);
		imp_sth->hstmt = NULL;
		return 0;
	}
	DBIc_on(imp_sth,DBIcf_IMPSET);

	return 1;
}

/******************************************************************************
***
***	Parameter:	Statement handle
***
***	Returnvalue:	Number of records affected.
***			-1 on statement which unknown record number
***			-2 on error
***
******************************************************************************/

int dbd_st_execute(SV *h, imp_sth_t *imp_sth)
{
	short n_columns;
	int rc,i;
	SWORD size;
	char colname[100];
	SWORD dummy;
	UDWORD uddummy;

	D_imp_dbh_from_sth;

	yard_log(1,"execute: hstmt = 0x%08x\n",imp_sth->hstmt);
	yard_clearerrorstack();

	rc = SQLExecute(imp_sth->hstmt);
	if (yard_clifailed(rc,imp_sth->hstmt,STMT_HANDLE))
		return -2;
	rc = SQLNumResultCols(imp_sth->hstmt,&n_columns);
	if (yard_clifailed(rc,imp_sth->hstmt,STMT_HANDLE))
	{
		SQLFreeStmt(imp_sth->hstmt,SQL_DROP);
		imp_sth->hstmt = NULL;
		return -2;
	}
	rc = SQLGetCursorName(imp_sth->hstmt,imp_sth->cursorname,sizeof(
		imp_sth->cursorname),&size);
	if (imp_sth->cursorname[0])
	{
		/*	SELECT or CALL				*/

		DBIc_NUM_FIELDS( imp_sth ) = n_columns;
		DBIc_on(imp_sth,DBIcf_ACTIVE);

		if ((imp_sth->cinfo = (struct col*)malloc(n_columns *
			sizeof(struct col))) == NULL)
		{
			SQLFreeStmt(imp_sth->hstmt,SQL_DROP);
			imp_sth->hstmt = NULL;
			return -2;
		}
		for (i=0;i<n_columns;i++)
		{
			rc = SQLDescribeCol(imp_sth->hstmt,i+1,
				imp_sth->cinfo[i].name,sizeof(
					imp_sth->cinfo[i].name), 
				&dummy,
				&imp_sth->cinfo[i].type,
				&uddummy,
				&dummy,
				&imp_sth->cinfo[i].nullable);
			if (yard_clifailed(rc,imp_sth->hstmt,STMT_HANDLE))
			{
				SQLFreeStmt(imp_sth->hstmt,SQL_DROP);
				imp_sth->hstmt = NULL;
				return(-2);
			}
			yard_log(2,"colname = %s, coltype = %d\n",
				imp_sth->cinfo[i].name,
				imp_sth->cinfo[i].type);
		}
	}

	return -1;
}

/******************************************************************************
***
***
***
******************************************************************************/

int dbd_st_bindsv(imp_sth_t *imp_sth, int idx, SV *val)
{
	int rc;
	int len;
	char *buf;

	buf = SvPV(val,len);
	yard_log(1,"bindsv: idx = %d (%s)\n",idx,buf);
	rc = SQLBindParameter(imp_sth->hstmt,idx,SQL_PARAM_INPUT,SQL_C_CHAR,
		SQL_CHAR,0,0,buf,len,NULL);
	if (yard_clifailed(rc,imp_sth->hstmt,STMT_HANDLE))
		return 0;
	return 1;
}

int dbd_st_rows(SV *sth)
{
	int rc;
	SDWORD n;

	D_imp_sth(sth);
	rc = SQLRowCount(imp_sth->hstmt,&n);
	if (yard_clifailed(rc,imp_sth->hstmt,STMT_HANDLE))
		return 0;
	return n;
}

/******************************************************************************
***
***
***
******************************************************************************/

int dbd_st_finish(imp_sth_t *imp_sth)
{
	int rc;

	yard_log(1,"st_finish: hstmt = 0x%08x\n", imp_sth->hstmt);

	DBIc_off(imp_sth,DBIcf_ACTIVE);

	rc = SQLFreeStmt(imp_sth->hstmt,SQL_CLOSE);
	if (yard_clifailed(rc,imp_sth->hstmt,STMT_HANDLE))
		return 0;
	return 1;
}

/******************************************************************************
***
***	Release a statement handle
***
******************************************************************************/

void dbd_st_destroy(imp_sth_t *imp_sth)
{
	int rc;

	yard_log(1,"st_destroy: hstmt = 0x%08x\n", imp_sth->hstmt);

	rc = SQLFreeStmt(imp_sth->hstmt,SQL_DROP);
	if (yard_clifailed(rc,imp_sth->hstmt,STMT_HANDLE))
		;
	if (imp_sth->cinfo)
		free(imp_sth->cinfo);
	imp_sth->hstmt = NULL;
	DBIc_off(imp_sth,DBIcf_IMPSET);
}

int dbd_st_STORE(SV *sth, SV *keysv, SV *valuesv)
{
	D_imp_sth(sth);
	STRLEN keyLength;
	char *key = SvPV( keysv, keyLength );
	SV *retsv = NULL;
	int on = SvTRUE(valuesv);

	yard_log(1,"st_STORE: key = %s\n", key);

	if (!strcmp(key,"LongTruncOk"))
	{
		yard_log(1,"Setting LongTruncOk to %d\n",on);
		if (on)
		{
			DBIc_on(imp_sth,DBIcf_LongTruncOk);
		}
		else
		{
			DBIc_off(imp_sth,DBIcf_LongTruncOk);
		}
		return TRUE;
	}
	return FALSE;
}

/******************************************************************************
***
***	Fetch a statement attribute.
***
******************************************************************************/

SV *dbd_st_FETCH(SV *sth, SV *keysv)
{
	D_imp_sth(sth);
	STRLEN keyLength;
	char *key = SvPV( keysv, keyLength );
	SV *retsv = NULL;

	yard_log(1,"st_FETCH: key = %s\n", key);

	if (!strcmp(key,"NUM_OF_FIELDS"))
	{
		retsv = newSViv((IV)DBIc_NUM_FIELDS(imp_sth));
	}
	if (!strcmp(key,"CursorName"))
	{
		retsv = newSVpv(imp_sth->cursorname,0);
	}
	if (!strcmp(key,"NAME"))
	{
		int i;
		AV *av = newAV();

		retsv = newRV((SV *)av);
		for (i = 0; i < DBIc_NUM_FIELDS( imp_sth ); i++)
		{
			av_store(av, i, newSVpv(imp_sth->cinfo[i].name, 0));
		}
	}
	if (!strcmp(key,"NULLABLE"))
	{
		int i;
		AV *av = newAV();

		retsv = newRV((SV *)av);
		for (i = 0; i < DBIc_NUM_FIELDS( imp_sth ); i++)
			av_store(av, i, newSViv(imp_sth->cinfo[i].nullable));
	}

	return ((retsv) ? sv_2mortal( retsv ) : Nullsv);
}


/******************************************************************************
***
***	Fetch the next row.
***
******************************************************************************/

AV *dbd_st_fetch(imp_sth_t *imp_sth)
{
	int i,rc;
	SDWORD size;
        AV *av;
	char buf[GETDATABUFSIZE];
	int extending = 0;
	int tsize;
	int maxsize = DBIc_LongReadLen(imp_sth);

	yard_log(1,"fetch: hstmt = %d, maxsize = %d\n",imp_sth->hstmt,
		maxsize);
	yard_clearerrorstack();

	rc = SQLFetch(imp_sth->hstmt);
	if (yard_clifailed(rc,imp_sth->hstmt,STMT_HANDLE))
		return Nullav; /* error handling ?!! */
	if (rc == SQL_NO_DATA_FOUND)
		return Nullav;
	av = DBIS->get_fbav(imp_sth);

	for (i=0;i<DBIc_NUM_FIELDS( imp_sth );i++)
	{
		SV *sv = AvARRAY(av)[i];

		maxsize = DBIc_LongReadLen(imp_sth);
		while ((rc=SQLGetData(imp_sth->hstmt,i+1,SQL_C_CHAR,buf,
			GETDATABUFSIZE,&size)) == SQL_SUCCESS_WITH_INFO)
		{
/*yard_log(1,"buf1 is >%s< %d\n",buf,size);*/
			if (ISBLOB(imp_sth->cinfo[i].type))
			{
				if (maxsize >= GETDATABUFSIZE)
					tsize = GETDATABUFSIZE;
				else
					tsize = maxsize;
				maxsize -= tsize;
			}
			else
				tsize = GETDATABUFSIZE;

			if (!extending)
				sv_setpvn(sv,buf,tsize);
			else
				sv_catpvn(sv,buf,tsize);
			extending = 1;
		}

		if (yard_clifailed(rc,imp_sth->hstmt,STMT_HANDLE))
			return Nullav;

		if (DBIc_is(imp_sth, DBIcf_ChopBlanks) && imp_sth->cinfo[i].type == SQL_CHAR)
		{
			ba_trim(buf);
		}

		if (size == SQL_NULL_DATA)
		{
			/*	NULL value				*/

			SvOK_off(sv);
		}
		else
		{
/*yard_log(1,"buf is >%s< %d\n",buf,size);*/
			if (ISBLOB(imp_sth->cinfo[i].type))
			{
				if (maxsize >= size % GETDATABUFSIZE)
					tsize = size % GETDATABUFSIZE;
				else
					tsize = maxsize;
				maxsize -= tsize;
			}
			else
				tsize = size % GETDATABUFSIZE;
			if (!extending)
				sv_setpvn(sv,buf,tsize);
			else
				sv_catpvn(sv,buf,tsize);
		}
	}
	return(av);
}

/******************************************************************************
***
***
***
******************************************************************************/

static void yard_log(int level, char *fmt,...)
{
	va_list args;
/*	char filename[400];*/
/*	FILE *logfile;*/

/*	sprintf(filename,"/tmp/yard-perl.log");*/
/*	sprintf(filename,"/dev/tty");*/

	if (DBIS->debug < level)
		return;

/*	if ((logfile = fopen(filename,"a+")) == NULL)*/
/*		return;*/

	va_start(args,fmt);
	vfprintf(stderr,fmt,args);
	va_end(args);
}

/******************************************************************************
***
***
***
******************************************************************************/

static void yard_clearerrorstack()
{
	int i;

	for (i=0;i<MAXERRORSTACKSIZE;i++)
		errorstack[i][0] = '\0';
	errorstackpos = 0;
}

/******************************************************************************
***
***
***
******************************************************************************/

static int yard_pusherror(char *s,char *sqlstate,int sqlcode)
{
	yard_log(1,"pusherror: pushing error msg %s\n",s);

	if (errorstackpos >= MAXERRORSTACKSIZE)
		return(1);	/* No more space on error stack	*/
	if (errorstackpos == 0)
	{
		sv_setiv(dbd_errnum, (IV)sqlcode);
		sv_setpv(dbd_state, sqlstate);
		sprintf(errorstack[errorstackpos++],"%s",s);
		sv_setpv(dbd_errstr, s);
	}
	else
	{
		sprintf(errorstack[errorstackpos],"%s (%d: %s)",
			errorstack[errorstackpos-1],sqlcode,s);
		sv_setpv(dbd_errstr, errorstack[errorstackpos]);
		errorstackpos++;
	}
	return(0);
}

/******************************************************************************
***
***	Fill the error stack when an error occured.
***
***	Return value:	0, no error occured
***			1, an error occured
***
******************************************************************************/

static int yard_clifailed(int rc, void *handle, int type)
{
	SWORD msglen;
	UDWORD sqlcode;
	char msg[ERRBUFSIZE+1];
	char sqlstate[ERRBUFSIZE+1];

	if (rc == SQL_ERROR)
	{
		yard_log(1,"clifailed: An error occured\n","");
		do
		{
			switch (type)
			{
				case ENV_HANDLE:
					rc = SQLError(handle,NULL,NULL,sqlstate,
					    &sqlcode,msg,ERRBUFSIZE,&msglen);
					break;
				case CON_HANDLE:
					rc = SQLError(NULL,handle,NULL,sqlstate,
					    &sqlcode,msg,ERRBUFSIZE,&msglen);
					break;
				case STMT_HANDLE:
					rc = SQLError(NULL,NULL,handle,sqlstate,
					    &sqlcode,msg,ERRBUFSIZE,&msglen);
					break;
				default:
					yard_log(1,"Invalid handle type %d\n",
						type);
			}

			switch (rc)
			{
				case SQL_SUCCESS:
				case SQL_SUCCESS_WITH_INFO:
					if (yard_pusherror(msg,sqlstate,sqlcode))
						return(1);
					break;
				case SQL_NO_DATA_FOUND:
					/*	No more information	*/

					return(1);
					break;
				case SQL_ERROR:
					if (yard_pusherror("error not found","XXXXX",-1))
						return(1);
					return(1);
				case SQL_INVALID_HANDLE:
					if (yard_pusherror("invalid handle","XXXXX",-1))
						return(1);
					return(1);
				default:
					yard_log(1,"Invalid rc %d\n", rc);
			}
		}
		while (1);
	}
	if (rc == SQL_INVALID_HANDLE)
	{
		if (yard_pusherror("invalid handle","XXXXX",-1))
			return(1);
		return(1);
	}
	return(0);
}
