/****************************************************************************\
*                                                                            *
* PHP/FI                                                                     *
*                                                                            *
* Copyright 1995,1996,1997 Rasmus Lerdorf                                    *
*                                                                            *
*  This program is free software; you can redistribute it and/or modify      *
*  it under the terms of the GNU General Public License as published by      *
*  the Free Software Foundation; either version 2 of the License, or         *
*  (at your option) any later version.                                       *
*                                                                            *
*  This program is distributed in the hope that it will be useful,           *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*  GNU General Public License for more details.                              *
*                                                                            *
*  You should have received a copy of the GNU General Public License         *
*  along with this program; if not, write to the Free Software               *
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
*                                                                            *
\****************************************************************************/

/******************************************************************************
***
***	YARD interface for PHP
***
******************************************************************************/

#include <stdarg.h>

#include "php.h"
#include "parse.h"
#include "clibasic.h"
#include "sql.h"
#include "sqlext.h"

#define	ERRBUFSIZE		512
#define	GETDATABUFSIZE		4096
#define	MAXERRORSTACKSIZE	10

#define ENV_HANDLE		0x001
#define CON_HANDLE		0x002
#define STMT_HANDLE		0x003

static char logfilename[1024];
static char envstring[1024];
static HENV henv = NULL;
static char errorstack[MAXERRORSTACKSIZE][ERRBUFSIZE+1+20];
static int errorstackpos;

static int yard_clifailed(int,void*,int);
static void *yard_str2handle(char *);
static char *yard_handle2str(void*);
static void yard_log(char *,...);
static int yard_pusherror(char*,char*,int);
static void yard_clearerrorstack();

/******************************************************************************
***
***	Initialisation function. This function must be called before any other
***	function is called.
***	If this function fails, additional error information might be 
***	displayed on stderr.
***
***	Parameter:	YARDHOME, the directory in which YARD is installed.
***			Logfilename or "" for no logging
***
***	Returnvalue:	"1" on success
***			"" on error
***
******************************************************************************/

void yard_init()
{
	Stack *s;
	int rc;

	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_init");
		Push("", STRING);
		return;
	}
	strcpy(logfilename,s->strval);

	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_init");
		Push("", STRING);
		return;
	}
	sprintf(envstring,"YARDHOME=%s",s->strval);
	putenv(envstring);

	yard_log("yard_init: YARDHOME = #%s#, logfile = #%s#\n",
		s->strval,logfilename);

	rc = SQLAllocEnv(&henv);
	if (rc)
	{
		yard_pusherror("SQLAllocEnv() failed","XXXXX",-1);
		Push("", STRING);
		return;
	}
	Push("1", STRING);
}

/******************************************************************************
***
***	Connect to a DBS/database
***
***	Parameter:	DBS (yarddbs@system)
***			userid
***			password
***
***	Returnvalue:	Connection handle as string on success
***			"" on error
***
******************************************************************************/

void yard_connect()
{
	int rc;
	char *dbs;
	char *userid;
	char *password;
	HDBC hdbc;
	Stack *s;
	char dsn[100];
	
	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_connect");
		Push("", STRING);
		return;
	}
	password = s->strval;

	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_connect");
		Push("", STRING);
		return;
	}
	userid = s->strval;

	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_connect");
		Push("", STRING);
		return;
	}
	dbs = s->strval;

	yard_log("yard_connect: dbs = #%s#, userid = #%s#, password %sspecified\n",dbs,userid,password[0]?"":"not ");
	yard_clearerrorstack();

	rc = SQLAllocConnect(henv,&hdbc);
	if (yard_clifailed(rc,henv,ENV_HANDLE))
	{
		Push("", STRING);
		return;
	}

	strcpy(dsn,dbs);
	rc = SQLConnect(hdbc,dsn,strlen(dsn),userid,strlen(userid),password,
		strlen(password));
	if (yard_clifailed(rc,hdbc,CON_HANDLE))
	{
		SQLFreeConnect(hdbc);
		Push("", STRING);
		return;
	}

	Push(yard_handle2str(hdbc), STRING);
}

/******************************************************************************
***
***	Information about the latest error that occured.
***
***	Parameter:	No
***
***	Returnvalue:	"", when no more error messages are available.
***			When the first call doesn't return anything, the last
***			function call has been succesful.
***			yard_error() may be called as long
***			as it returns an error string in order to get all
***			messages associated with the last error.
***
******************************************************************************/

void yard_error()
{
	int i;

	for (i=0;i<MAXERRORSTACKSIZE;i++)
		if (errorstack[i][0])
		{
			Push(errorstack[i], STRING);
			errorstack[i][0] = '\0';
			return;
		}
	Push("", STRING);
}
	
/******************************************************************************
***
***
***
******************************************************************************/

static void yard_clearerrorstack()
{
	int i;

	for (i=0;i<MAXERRORSTACKSIZE;i++)
		errorstack[i][0] = '\0';
	errorstackpos = 0;
}

/******************************************************************************
***
***	Parameter:	Connection handle
***			SQL Statement
***
***	Returnvalue:	Statement handle as string on success
***			"" on error
***
******************************************************************************/

void yard_prepare()
{
	Stack *s;
	char *stmt;
	HDBC hdbc;
	HSTMT hstmt;
	int rc;
	
	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_prepare");
		Push("", STRING);
		return;
	}
	stmt = s->strval;

	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_prepare");
		Push("", STRING);
		return;
	}
	hdbc = yard_str2handle(s->strval);

	yard_log("yard_prepare: stmt = #%s#\n",stmt);
	yard_clearerrorstack();

	rc = SQLAllocStmt(hdbc,&hstmt);
	if (yard_clifailed(rc,hdbc,CON_HANDLE))
	{
		Push("", STRING);
		return;
	}
	rc = SQLPrepare(hstmt,stmt,strlen(stmt));
	if (yard_clifailed(rc,hstmt,STMT_HANDLE))
	{
		SQLFreeStmt(hstmt,SQL_DROP);
		Push("", STRING);
		return;
	}

	Push(yard_handle2str(hstmt), STRING);
}

/******************************************************************************
***
***	Parameter:	Statement handle
***
***	Returnvalue:	
***			"1" on successful execution.
***			"" on error
***
******************************************************************************/

void yard_execute()
{
	Stack *s;
	int rc;
	HSTMT hstmt;
	
	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_execute");
		Push("", STRING);
		return;
	}
	hstmt = yard_str2handle(s->strval);

	yard_log("yard_execute: hstmt = %d\n",hstmt);
	yard_clearerrorstack();

	rc = SQLExecute(hstmt);
	if (yard_clifailed(rc,hstmt,STMT_HANDLE))
	{
		Push("", STRING);
		return;
	}
	Push("1", STRING);
}

/******************************************************************************
***
***	Parameter:	Statement handle
***			column index, starting at 1
***
***	Returnvalue:	Column value of the current row in the Resultset on
***				success.
***			"" on error. Be sure to call yard_error() when ""
***				is returned for distinguishing an empty string
***				and an error.
***
******************************************************************************/

void yard_getdata()
{
	Stack *s;
	int rc,index;
	HSTMT hstmt;
	char *buf,*p;
	UDWORD size,totalsize;
	
	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_getdata");
		Push("", STRING);
		return;
	}
	index = s->intval;

	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_getdata");
		Push("", STRING);
		return;
	}
	hstmt = yard_str2handle(s->strval);

	yard_log("yard_getdata: hstmt = %d\n",hstmt);
	yard_clearerrorstack();

	/**
	***	Special treatment for "long" columns like BLOBs
	**/

	p = buf = (char*)emalloc(2,GETDATABUFSIZE+1);
	if (!p)
	{
		Error("Out of memory");
		Push("", STRING);
		return;
	}
	totalsize = 0;
	size = GETDATABUFSIZE;
	while ((rc=SQLGetData(hstmt,index,SQL_C_CHAR,buf+totalsize,
		size,&size)) == SQL_SUCCESS_WITH_INFO)
	{
		SWORD msglen;
		UDWORD sqlcode;
		char msg[ERRBUFSIZE+1];
		char sqlstate[ERRBUFSIZE+1];

		rc = SQLError(NULL,NULL,hstmt,sqlstate,
			&sqlcode,msg,ERRBUFSIZE,&msglen);
		if (strcmp(sqlstate,"01004"))
		{
			Error("long data error");
			Push("", STRING);
			return;
		}

		/**
		***	<size> is set to the total number of bytes available
		**/

		totalsize = GETDATABUFSIZE;
		buf = emalloc(2, size + 1);
		if (!buf)
		{
			Error("Out of memory");
			Push("", STRING);
			return;
		}
		size -= GETDATABUFSIZE;
	}

	if (yard_clifailed(rc,hstmt,STMT_HANDLE))
	{
		Push("", STRING);
		return;
	}
	Push(buf, STRING);
}

/******************************************************************************
***
***	Parameter:	Statement handle
***			Direction (for future use)
***
***	Returnvalue:	"100" when the end of Resultset is reached.
***			"0" on success.
***			"" on error.
***
******************************************************************************/

void yard_fetch()
{
	Stack *s;
	int rc;
	HSTMT hstmt;
	
	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_fetch");
		Push("", STRING);
		return;
	}
	if ((s = Pop()) == NULL)
	{
		Error("Stack error in yard_fetch");
		Push("", STRING);
		return;
	}
	hstmt = yard_str2handle(s->strval);

	yard_log("yard_fetch: hstmt = %d\n",hstmt);
	yard_clearerrorstack();

	rc = SQLFetch(hstmt);
	if (yard_clifailed(rc,hstmt,STMT_HANDLE))
	{
		Push("", STRING);
		return;
	}
	if (rc == SQL_NO_DATA_FOUND)
	{
		Push("100", STRING);
		return;
	}
	Push("0", STRING);
}

/******************************************************************************
***
***	Fill the error stack when an error occured.
***
***	Return value:	0, no error occured
***			1, an error occured
***
******************************************************************************/

static int yard_clifailed(int rc, void *handle, int type)
{
	SWORD msglen;
	UDWORD sqlcode;
	char msg[ERRBUFSIZE+1];
	char sqlstate[ERRBUFSIZE+1];

	if (rc == SQL_ERROR)
	{
		yard_log("clifailed: An error occured\n");
		do
		{
			switch (type)
			{
				case ENV_HANDLE:
					rc = SQLError(handle,NULL,NULL,sqlstate,
					    &sqlcode,msg,ERRBUFSIZE,&msglen);
					break;
				case CON_HANDLE:
					rc = SQLError(NULL,handle,NULL,sqlstate,
					    &sqlcode,msg,ERRBUFSIZE,&msglen);
					break;
				case STMT_HANDLE:
					rc = SQLError(NULL,NULL,handle,sqlstate,
					    &sqlcode,msg,ERRBUFSIZE,&msglen);
					break;
			}

			switch (rc)
			{
				case SQL_SUCCESS:
				case SQL_SUCCESS_WITH_INFO:
					if (yard_pusherror(msg,sqlstate,sqlcode))
						return(1);
					break;
				case SQL_NO_DATA_FOUND:
					/*	No more information	*/

					return(1);
					break;
				case SQL_ERROR:
					if (yard_pusherror("error not found","XXXXX",-1))
						return(1);
					return(1);
				case SQL_INVALID_HANDLE:
					if (yard_pusherror("invalid handle","XXXXX",-1))
						return(1);
					return(1);
			}
		}
		while (1);
	}
	if (rc == SQL_INVALID_HANDLE)
	{
		if (yard_pusherror("invalid handle","XXXXX",-1))
			return(1);
		return(1);
	}
	return(0);
}

/******************************************************************************
***
***
***
******************************************************************************/

static int yard_pusherror(char *s,char *sqlstate,int sqlcode)
{
	yard_log("pusherror: pushing error msg %s\n",s);

	if (errorstackpos >= MAXERRORSTACKSIZE)
		return(1);	/* No more space on error stack	*/
	sprintf(errorstack[errorstackpos++],"%s:%d:%s",sqlstate,sqlcode,s);
	return(0);
}

/******************************************************************************
***
***
***
******************************************************************************/

static void *yard_str2handle(char *s)
{
	void *p;

	p = (void*)strtoul(s,NULL,16);

	return(p);
}

/******************************************************************************
***
***
***
******************************************************************************/

static char *yard_handle2str(void *p)
{
	static char tmp[100];

	sprintf(tmp,"0x%x",p);
	return(tmp);
}

/******************************************************************************
***
***
***
******************************************************************************/

static void yard_log(char *fmt,...)
{
	va_list args;
	FILE *fp;

	if (!logfilename[0])
		return;

	if ((fp = fopen(logfilename,"a")) == NULL)
		return;

	va_start(args,fmt);
	vfprintf(fp,fmt,args);
	va_end(args);
	fclose(fp);
}
